@@ -1,5 +1,5 @@
 from pathlib import Path
-from typing import TYPE_CHECKING, Dict, List, Sequence
+from typing import TYPE_CHECKING, Dict, List, Sequence, Union
 
 import matplotlib.pyplot as plt
 import numpy as np
@@ -10,6 +10,8 @@ from sklearn.preprocessing import LabelEncoder
 
 from eir.experiment_io.label_transformer_io import load_transformers
 from eir.interpretation.interpretation_utils import (
+    TargetTypeInfo,
+    get_appropriate_target_transformer,
     get_long_format_attribution_df,
     plot_attributions_bar,
     stratify_attributions_by_target_classes,
@@ -19,6 +21,9 @@ from eir.predict_modules.predict_tabular_input_setup import (
     ComputedPredictTabularInputInfo,
 )
 from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
+from eir.setup.output_setup_modules.survival_output_setup import (
+    ComputedSurvivalOutputInfo,
+)
 from eir.setup.output_setup_modules.tabular_output_setup import (
     ComputedTabularOutputInfo,
 )
@@ -27,6 +32,10 @@ from eir.utils.logging import get_logger
 
 if TYPE_CHECKING:
     from eir.interpretation.interpretation import SampleAttribution
+    from eir.predict import PredictExperiment
+    from eir.predict_modules.predict_attributions import (
+        LoadedTrainExperimentMixedWithPredict,
+    )
     from eir.train import Experiment
 
 
@@ -34,11 +43,14 @@ logger = get_logger(__name__)
 
 
 def analyze_tabular_input_attributions(
-    experiment: "Experiment",
+    experiment: Union[
+        "Experiment",
+        "PredictExperiment",
+        "LoadedTrainExperimentMixedWithPredict",
+    ],
     input_name: str,
     output_name: str,
-    target_column_name: str,
-    target_column_type: str,
+    target_info: TargetTypeInfo,
     attribution_outfolder: Path,
     all_attributions: Sequence["SampleAttribution"],
 ):
@@ -85,15 +97,24 @@ def analyze_tabular_input_attributions(
     df_attributions.to_csv(attribution_outfolder / "feature_importances.csv")
 
     output_object = exp.outputs[output_name]
-    assert isinstance(output_object, ComputedTabularOutputInfo)
-    target_transformer = output_object.target_transformers[target_column_name]
+    assert isinstance(
+        output_object, (ComputedTabularOutputInfo, ComputedSurvivalOutputInfo)
+    )
+
+    target_column_name = target_info.name
+    target_column_type = target_info.type_
+
+    target_transformer = get_appropriate_target_transformer(
+        output_object=output_object,
+        target_column_name=target_column_name,
+        target_column_type=target_column_type,
+    )
 
     all_attributions_class_stratified = stratify_attributions_by_target_classes(
         all_attributions=all_attributions,
         target_transformer=target_transformer,
         output_name=output_name,
-        target_column=target_column_name,
-        column_type=target_column_type,
+        target_info=target_info,
     )
 
     for class_name, class_attributions in all_attributions_class_stratified.items():
