@@ -48,18 +48,27 @@ from eir.interpretation.interpret_array import (
 from eir.interpretation.interpret_image import analyze_image_input_attributions
 from eir.interpretation.interpret_omics import (
     OmicsConsumerCallable,
+    ParsedOmicsAttributions,
     analyze_omics_input_attributions,
     get_omics_consumer,
 )
 from eir.interpretation.interpret_sequence import analyze_sequence_input_attributions
 from eir.interpretation.interpret_tabular import analyze_tabular_input_attributions
-from eir.interpretation.interpretation_utils import MyIntegratedGradients
+from eir.interpretation.interpretation_utils import (
+    MyIntegratedGradients,
+    TargetTypeInfo,
+    get_appropriate_target_transformer,
+)
 from eir.models.input.omics.omics_models import CNNModel, LinearModel
 from eir.models.model_training_utils import gather_data_loader_samples
+from eir.setup.output_setup_modules.survival_output_setup import (
+    ComputedSurvivalOutputInfo,
+)
 from eir.setup.output_setup_modules.tabular_output_setup import (
     ComputedTabularOutputInfo,
+    TabularOutputTypeConfig,
 )
-from eir.setup.schemas import InputConfig, OutputConfig
+from eir.setup.schemas import InputConfig, OutputConfig, SurvivalOutputTypeConfig
 from eir.target_setup.target_label_setup import MissingTargetsInfo
 from eir.train_utils.evaluation import validation_handler
 from eir.train_utils.utils import (
@@ -221,7 +230,9 @@ def get_background_loader(experiment: "Experiment") -> torch.utils.data.DataLoad
 def tabular_attribution_analysis_wrapper(
     model: "al_meta_model",
     experiment: Union[
-        "Experiment", "PredictExperiment", "LoadedTrainExperimentMixedWithPredict"
+        "Experiment",
+        "PredictExperiment",
+        "LoadedTrainExperimentMixedWithPredict",
     ],
     output_folder_target_callable: Callable,
     dataset_to_interpret: al_local_datasets,
@@ -242,9 +253,20 @@ def tabular_attribution_analysis_wrapper(
     for output_name, target_column_type, target_column_name in target_columns_gen:
         output_type = exp.outputs[output_name].output_config.output_info.output_type
 
-        if output_type != "tabular":
+        if output_type not in ("tabular", "survival"):
             continue
 
+        if output_type == "survival":
+            output_type_info = exp.outputs[output_name].output_config.output_type_info
+            assert isinstance(output_type_info, SurvivalOutputTypeConfig)
+            if target_column_name == output_type_info.time_column:
+                continue
+
+            if output_type_info.loss_function == "CoxPHLoss":
+                target_column_type = "con"
+            else:
+                target_column_type = "cat"
+
         ao = get_attribution_object(
             experiment=exp,
             model=model_copy,
@@ -289,19 +311,37 @@ def tabular_attribution_analysis_wrapper(
         )
 
         output_object = exp.outputs[output_name]
-        assert isinstance(output_object, ComputedTabularOutputInfo)
-        target_transformer = output_object.target_transformers[target_column_name]
+        assert isinstance(
+            output_object, (ComputedTabularOutputInfo, ComputedSurvivalOutputInfo)
+        )
+
+        target_transformer = get_appropriate_target_transformer(
+            output_object=output_object,
+            target_column_name=target_column_name,
+            target_column_type=target_column_type,
+        )
+
+        is_bce = False
+        output_type_info = output_object.output_config.output_type_info
+        if isinstance(output_type_info, TabularOutputTypeConfig):
+            is_bce = output_type_info.cat_loss_name == "BCEWithLogitsLoss"
+
+        target_column_info = TargetTypeInfo(
+            name=target_column_name,
+            type_=target_column_type,
+            cat_is_bce=is_bce,
+        )
 
         act_consumers = get_attribution_consumers(
             input_names_and_types=input_names_and_types,
             target_transformer=target_transformer,
             output_name=output_name,
-            target_column=target_column_name,
-            column_type=target_column_type,
+            target_info=target_column_info,
         )
 
         all_attributions = process_attributions_for_all_modalities(
-            attribution_producer=act_producer, attribution_consumers=act_consumers
+            attribution_producer=act_producer,
+            attribution_consumers=act_consumers,
         )
 
         for input_name in ao.input_names_ordered:
@@ -320,21 +360,28 @@ def tabular_attribution_analysis_wrapper(
                 input_name=input_name,
                 output_name=output_name,
             )
-            common_kwargs = {
-                "experiment": experiment,
-                "input_name": input_name,
-                "target_column_name": target_column_name,
-                "target_column_type": target_column_type,
-                "all_attributions": all_attributions[input_name],
-                "attribution_outfolder": act_output_folder,
-            }
+
+            cur_attributions = all_attributions[input_name]
 
             if input_type == "omics":
-                analyze_omics_input_attributions(**common_kwargs)
+                assert isinstance(cur_attributions, ParsedOmicsAttributions)
+                analyze_omics_input_attributions(
+                    experiment=experiment,
+                    input_name=input_name,
+                    all_attributions=cur_attributions,
+                    target_info=target_column_info,
+                    attribution_outfolder=act_output_folder,
+                )
 
             elif input_type == "tabular":
+                assert isinstance(cur_attributions, Sequence)
                 analyze_tabular_input_attributions(
-                    **common_kwargs, output_name=output_name
+                    experiment=experiment,
+                    input_name=input_name,
+                    target_info=target_column_info,
+                    all_attributions=cur_attributions,
+                    attribution_outfolder=act_output_folder,
+                    output_name=output_name,
                 )
 
             elif input_type == "sequence":
@@ -345,22 +392,32 @@ def tabular_attribution_analysis_wrapper(
                     model=model_copy,
                     baselines=ao.baselines_with_extras,
                 )
+                assert isinstance(cur_attributions, Sequence)
                 analyze_sequence_input_attributions(
-                    **common_kwargs,
+                    experiment=experiment,
+                    input_name=input_name,
+                    target_info=target_column_info,
+                    all_attributions=cur_attributions,
+                    attribution_outfolder=act_output_folder,
                     expected_target_classes_attributions=expected_value,
                     output_name=output_name,
                 )
             elif input_type == "image":
+                assert isinstance(cur_attributions, Sequence)
                 analyze_image_input_attributions(
-                    **common_kwargs, output_name=output_name
+                    experiment=experiment,
+                    input_name=input_name,
+                    target_info=target_column_info,
+                    all_attributions=cur_attributions,
+                    attribution_outfolder=act_output_folder,
+                    output_name=output_name,
                 )
 
             elif input_type == "array":
-                cur_array_attributions = all_attributions[input_name]
-                assert isinstance(cur_array_attributions, dict)
+                assert isinstance(cur_attributions, dict)
                 analyze_array_input_attributions(
                     attribution_outfolder=act_output_folder,
-                    all_attributions=cur_array_attributions,
+                    all_attributions=cur_attributions,
                 )
 
         ao.hook_handle.remove()
@@ -596,10 +653,14 @@ def get_attribution_consumers(
     ],
     target_transformer: "al_label_transformers_object",
     output_name: str,
-    target_column: str,
-    column_type: str,
-) -> Dict[
-    str, Union["BasicConsumerCallable", OmicsConsumerCallable, ArrayConsumerCallable]
+    target_info: TargetTypeInfo,
+) -> dict[
+    str,
+    Union[
+        "BasicConsumerCallable",
+        OmicsConsumerCallable,
+        ArrayConsumerCallable,
+    ],
 ]:
     consumers_dict = {}
 
@@ -612,8 +673,7 @@ def get_attribution_consumers(
             input_name=input_name,
             target_transformer=target_transformer,
             output_name=output_name,
-            target_column=target_column,
-            column_type=column_type,
+            target_info=target_info,
         )
         if consumer is not None:
             consumers_dict[input_name] = consumer
@@ -633,8 +693,7 @@ def _get_consumer_from_input_type(
     input_type: str,
     input_name: str,
     output_name: str,
-    target_column: str,
-    column_type: str,
+    target_info: TargetTypeInfo,
 ) -> BasicConsumerCallable | OmicsConsumerCallable | ArrayConsumerCallable:
     if input_type in ("sequence", "tabular", "image"):
         return get_basic_consumer()
@@ -644,8 +703,7 @@ def _get_consumer_from_input_type(
             target_transformer=target_transformer,
             input_name=input_name,
             output_name=output_name,
-            target_column=target_column,
-            column_type=column_type,
+            target_info=target_info,
         )
 
     elif input_type == "array":
@@ -653,8 +711,7 @@ def _get_consumer_from_input_type(
             target_transformer=target_transformer,
             input_name=input_name,
             output_name=output_name,
-            target_column=target_column,
-            column_type=column_type,
+            target_info=target_info,
         )
     else:
         raise ValueError(f"Unsupported input type {input_type}.")
@@ -670,15 +727,12 @@ def get_sample_attribution_producer(
     missing_for_modality = missing_ids_per_output.missing_ids_per_modality
     cur_missing_ids = missing_for_modality[output_name]
 
-    missing_within_modality = missing_ids_per_output.missing_ids_within_modality
-    cur_missing_ids_within = missing_within_modality[output_name][target_column_name]
-
     for batch, raw_inputs in data_producer:
         sample_target_labels = batch.target_labels
         cur_target_label = sample_target_labels[output_name][target_column_name]
 
         cur_id = batch.ids[0]
-        if cur_id in cur_missing_ids or cur_id in cur_missing_ids_within:
+        if cur_id in cur_missing_ids:
             continue
 
         sample_all_modalities_attributions = act_func(
@@ -716,7 +770,9 @@ def get_basic_consumer() -> BasicConsumerCallable:
 def process_attributions_for_all_modalities(
     attribution_producer: Generator["SampleAttribution", None, None],
     attribution_consumers: Dict[str, Callable],
-) -> dict[str, Sequence["SampleAttribution"] | dict[str, np.ndarray]]:
+) -> dict[
+    str, Sequence["SampleAttribution"] | dict[str, np.ndarray] | ParsedOmicsAttributions
+]:
     processed_attributions_all_modalities = {}
 
     for sample_attribution in attribution_producer:
@@ -843,7 +899,9 @@ def get_attributions(
 
 def _get_interpretation_data_producer(
     experiment: Union[
-        "Experiment", "PredictExperiment", "LoadedTrainExperimentMixedWithPredict"
+        "Experiment",
+        "PredictExperiment",
+        "LoadedTrainExperimentMixedWithPredict",
     ],
     column_name: str,
     column_type: str,
@@ -851,7 +909,9 @@ def _get_interpretation_data_producer(
     dataset: al_local_datasets,
 ) -> Generator[Tuple[Batch, dict[str, Any]], None, None]:
     output_object = experiment.outputs[output_name]
-    assert isinstance(output_object, ComputedTabularOutputInfo)
+    assert isinstance(
+        output_object, (ComputedTabularOutputInfo, ComputedSurvivalOutputInfo)
+    )
 
     target_transformer = output_object.target_transformers[column_name]
     gc = experiment.configs.global_config
@@ -879,10 +939,41 @@ def _get_interpretation_data_producer(
 
         batch = state["batch"]
 
+        target_labels = batch.target_labels
+        match (output_object, column_type):
+            case (ComputedSurvivalOutputInfo(), "cat"):
+                # here we inject the *time* target label as the event target label
+                # as we will be looking at attributions towards each output node
+                # (time bin) in the case of discrete survival outputs
+                output_type_info = output_object.output_config.output_type_info
+                assert isinstance(output_type_info, SurvivalOutputTypeConfig)
+                time_column = output_type_info.time_column
+                event_column = output_type_info.event_column
+                time_bin = target_labels[output_name][time_column].to(torch.long)
+                target_labels[output_name][event_column] = time_bin
+            case (ComputedTabularOutputInfo(), "cat"):
+                output_type_info = output_object.output_config.output_type_info
+                assert isinstance(output_type_info, TabularOutputTypeConfig)
+                if output_type_info.cat_loss_name == "BCEWithLogitsLoss":
+                    # here the network only produces 1 output per target
+                    # so we inject the 0 here to get the attributions
+                    # for the positive class
+                    prev_label = target_labels[output_name][column_name]
+                    new_label = torch.zeros_like(prev_label)
+                    target_labels[output_name][column_name] = new_label
+
+        cur_label = target_labels[output_name][column_name]
+        cur_column_type = column_type
+
+        if cur_column_type == "con" and torch.isnan(cur_label).any():
+            continue
+        elif cur_column_type == "cat" and cur_label < 0:
+            continue
+
         inputs_detached = _detach_all_inputs(tensor_inputs=batch.inputs)
         batch_interpretation = Batch(
             inputs=inputs_detached,
-            target_labels=batch.target_labels,
+            target_labels=target_labels,
             ids=batch.ids,
         )
 
@@ -941,9 +1032,9 @@ def _get_categorical_sample_indices_for_attributions(
     acc_label_limit = max_attributions_per_class
     indices = []
 
-    for index, sample in enumerate(dataset.samples):
-        target_labels = sample.target_labels
-        cur_sample_target_label = target_labels[output_name][target_column]
+    for idx in range(len(dataset)):
+        row = dataset.target_labels_df.row(idx, named=True)
+        cur_sample_target_label = row[f"{output_name}__{target_column}"]
 
         is_over_limit = acc_label_counts[cur_sample_target_label] == acc_label_limit
         is_not_in_target_classes = (
@@ -952,7 +1043,7 @@ def _get_categorical_sample_indices_for_attributions(
         if is_over_limit or is_not_in_target_classes:
             continue
 
-        indices.append(index)
+        indices.append(idx)
         acc_label_counts[cur_sample_target_label] += 1
 
     return tuple(indices)
