@@ -1,6 +1,7 @@
 import logging
 import random
 from collections import defaultdict
+from dataclasses import dataclass
 from pathlib import Path
 from typing import (
     TYPE_CHECKING,
@@ -30,9 +31,17 @@ from captum.attr._utils.approximation_methods import approximation_parameters
 from captum.attr._utils.common import _reshape_and_sum
 from matplotlib import pyplot as plt
 from scipy.stats import bootstrap
+from sklearn.preprocessing import KBinsDiscretizer
 from torch import Tensor
 
 from eir.setup import schemas
+from eir.setup.output_setup_modules.survival_output_setup import (
+    ComputedSurvivalOutputInfo,
+    SurvivalOutputTypeConfig,
+)
+from eir.setup.output_setup_modules.tabular_output_setup import (
+    ComputedTabularOutputInfo,
+)
 from eir.utils.logging import get_logger
 
 if TYPE_CHECKING:
@@ -43,38 +52,92 @@ if TYPE_CHECKING:
 logger = get_logger(name=__name__)
 
 
+@dataclass()
+class TargetTypeInfo:
+    name: str
+    type_: str
+    cat_is_bce: bool
+
+
 def get_target_class_name(
     sample_label: torch.Tensor,
     target_transformer: "al_label_transformers_object",
-    column_type: str,
-    target_column_name: str,
-):
-    if column_type == "con":
-        return target_column_name
+    target_info: TargetTypeInfo,
+) -> str:
+    """
+    Note that we have a special case when working with BCE, where the
+    mapping of index -> class name breaks down. In this case, we are always looking
+    at index 0 in the produced network outputs (i.e., sample label is patched
+    to be always 0) which noes not represent class_ 0 in the transformer,
+    but rather score towards the positive class (index 1).
+    """
+
+    if target_info.type_ == "con":
+        return target_info.name
 
     tt_it = target_transformer.inverse_transform
-    cur_trn_label = tt_it([sample_label.item()])[0]
+
+    match target_transformer:
+        case KBinsDiscretizer():
+            bin_idx = sample_label.item()
+
+            bin_edges = target_transformer.bin_edges_[0]
+
+            lower_bound = bin_edges[bin_idx]
+            upper_bound = bin_edges[bin_idx + 1]
+
+            return f"time_{lower_bound:.1f}_to_{upper_bound:.1f}"
+        case _:
+            if not target_info.cat_is_bce:
+                cur_trn_label = tt_it([sample_label.item()])[0]
+            else:
+                cur_trn_label = target_transformer.classes_[1]
 
     return cur_trn_label
 
 
+def get_appropriate_target_transformer(
+    output_object: Union[ComputedTabularOutputInfo, ComputedSurvivalOutputInfo],
+    target_column_name: str,
+    target_column_type: str,
+) -> "al_label_transformers_object":
+    """
+    We have this so we can inject the time binning transformer for discrete
+    survival outputs. Note that this is coupled with other injections done for the
+    interpretation of discrete survival outputs (e.g. injecting the current
+    time bin as an event target, which then correctly indexes into a given
+    time bin node in the model output).
+    """
+    target_transformer = output_object.target_transformers[target_column_name]
+
+    match output_object:
+        case ComputedSurvivalOutputInfo():
+            output_type_info = output_object.output_config.output_type_info
+            assert isinstance(output_type_info, SurvivalOutputTypeConfig)
+
+            if target_column_type == "cat":
+                target_transformer = output_object.target_transformers[
+                    output_type_info.time_column
+                ]
+
+    return target_transformer
+
+
 def stratify_attributions_by_target_classes(
     all_attributions: Sequence["SampleAttribution"],
     target_transformer: "al_label_transformers_object",
     output_name: str,
-    target_column: str,
-    column_type: str,
+    target_info: TargetTypeInfo,
 ) -> DefaultDict[Any, list["SampleAttribution"]]:
     all_attributions_target_class_stratified = defaultdict(list)
 
     for sample in all_attributions:
         cur_labels_all = sample.sample_info.target_labels
-        cur_labels = cur_labels_all[output_name][target_column]
+        cur_labels = cur_labels_all[output_name][target_info.name]
         cur_label_name = get_target_class_name(
             sample_label=cur_labels,
             target_transformer=target_transformer,
-            column_type=column_type,
-            target_column_name=target_column,
+            target_info=target_info,
         )
         all_attributions_target_class_stratified[cur_label_name].append(sample)
 
@@ -216,11 +279,13 @@ def calculate_top_n_tokens(
 
 def get_basic_sample_attributions_to_analyse_generator(
     interpretation_config: schemas.BasicInterpretationConfig,
-    all_attributions: list["SampleAttribution"],
+    all_attributions: Sequence["SampleAttribution"],
 ) -> Generator["SampleAttribution", None, None]:
     strategy = interpretation_config.interpretation_sampling_strategy
     n_samples = interpretation_config.num_samples_to_interpret
 
+    all_attributions = list(all_attributions)
+
     if strategy == "first_n":
         base = all_attributions[:n_samples]
     elif strategy == "random_sample":
