@@ -3,6 +3,7 @@ import os
 from collections import OrderedDict
 from copy import deepcopy
 from dataclasses import dataclass
+from functools import partial
 from pathlib import Path
 from typing import (
     Any,
@@ -214,7 +215,6 @@ def init_vocab(
     tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
 ) -> Vocab:
     if tokenizer_name == "bpe":
-        assert gathered_stats.total_count > 0
         tokenizer_object = extract_tokenizer_object_from_function(
             tokenizer_callable=tokenizer
         )
@@ -431,10 +431,57 @@ def get_bpe_tokenizer(
         return cast(TokenizerProtocolPreSplit, _tokenize_pre_split)
 
 
+class TokenizerVocabSizeError(Exception):
+    pass
+
+
+@dataclass
+class TokenizerValidationResult:
+    is_valid: bool
+    actual_vocab_size: int
+    requested_vocab_size: Optional[int]
+    error_message: Optional[str] = None
+
+
+def validate_tokenizer_vocab_size(
+    tokenizer: Tokenizer,
+    requested_vocab_size: Optional[int],
+) -> TokenizerValidationResult:
+    if requested_vocab_size is None:
+        return TokenizerValidationResult(
+            is_valid=True,
+            actual_vocab_size=tokenizer.get_vocab_size(),
+            requested_vocab_size=None,
+        )
+
+    actual_size = tokenizer.get_vocab_size()
+
+    if actual_size > requested_vocab_size:
+        return TokenizerValidationResult(
+            is_valid=False,
+            actual_vocab_size=actual_size,
+            requested_vocab_size=requested_vocab_size,
+            error_message=(
+                f"The final vocabulary size ({actual_size}) exceeds the requested "
+                f"maximum size ({requested_vocab_size}). This can happen when the "
+                f"requested vocabulary size is too small for the complexity of your "
+                f"training data. Please increase the vocab_size parameter to at least "
+                f"{actual_size} tokens."
+            ),
+        )
+
+    return TokenizerValidationResult(
+        is_valid=True,
+        actual_vocab_size=actual_size,
+        requested_vocab_size=requested_vocab_size,
+    )
+
+
 def _get_bpe_tokenizer_object(
     vocab_iterator: Optional[Iterator],
     vocab_file: Optional[str],
     vocab_size: Optional[int],
+    raise_on_validation_error: bool = True,
 ) -> Tokenizer:
     if vocab_file:
         logger.info("Loading BPE vocabulary from file %s.", vocab_file)
@@ -447,7 +494,8 @@ def _get_bpe_tokenizer_object(
 
         tokenizer = Tokenizer.from_file(path=vocab_file)
     else:
-        assert vocab_iterator is not None
+        if vocab_iterator is None:
+            raise ValueError("Either vocab_iterator or vocab_file must be provided.")
 
         logger.info("Training BPE tokenizer from source data.")
 
@@ -466,18 +514,44 @@ def _get_bpe_tokenizer_object(
 
         tokenizer.train_from_iterator(iterator=vocab_iterator, trainer=trainer)
 
+        validation_result = validate_tokenizer_vocab_size(
+            tokenizer=tokenizer,
+            requested_vocab_size=vocab_size,
+        )
+
+        if not validation_result.is_valid:
+            logger.warning(validation_result.error_message)
+
+            if raise_on_validation_error:
+                raise TokenizerVocabSizeError(validation_result.error_message)
+
     return tokenizer
 
 
+def identity_tokenize(raw_input_split: Sequence[str]) -> Sequence[str]:
+    return raw_input_split
+
+
+def join_and_tokenize(
+    raw_input_split: Sequence[str], base_tokenizer: Any
+) -> Sequence[str]:
+    input_joined = " ".join(raw_input_split)
+    return base_tokenizer(input_joined)
+
+
+def make_join_tokenizer(base_tokenizer: Any) -> TokenizerProtocolPreSplit:
+    return cast(
+        TokenizerProtocolPreSplit,
+        partial(join_and_tokenize, base_tokenizer=base_tokenizer),
+    )
+
+
 def get_basic_tokenizer(
     tokenizer_name: al_tokenizer_choices,  # type: ignore
     tokenizer_language: Optional[str],
 ) -> TokenizerProtocolPreSplit:
-    def _identity(raw_input_split: Sequence[str]) -> Sequence[str]:
-        return raw_input_split
-
     if not tokenizer_name:
-        return _identity
+        return identity_tokenize
 
     _validate_pytorch_tokenizer_args(
         tokenizer_name=tokenizer_name, tokenizer_language=tokenizer_language
@@ -486,19 +560,11 @@ def get_basic_tokenizer(
         "Using tokenizer '%s' with language '%s'.", tokenizer_name, tokenizer_language
     )
 
-    tokenizer = get_pytorch_tokenizer(
+    base_tokenizer = get_pytorch_tokenizer(
         tokenizer=tokenizer_name, language=tokenizer_language
     )
 
-    def _join_and_tokenize(raw_input_split: Sequence[str]) -> Sequence[str]:
-        """
-        " ".join(...) since the PyTorch tokenizers operate on a single string.
-        Can be optimized later by not splitting the input in the first place.
-        """
-        input_joined = " ".join(raw_input_split)
-        return tokenizer(input_joined)
-
-    return cast(TokenizerProtocolPreSplit, _join_and_tokenize)
+    return cast(TokenizerProtocolPreSplit, make_join_tokenizer(base_tokenizer))
 
 
 def _validate_pytorch_tokenizer_args(
@@ -520,22 +586,40 @@ def _validate_pytorch_tokenizer_args(
         )
 
 
+@overload
+def encode_text(
+    raw_input: str,
+    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
+    pytorch_vocab: Vocab,
+) -> Sequence[int]: ...
+
+
+@overload
+def encode_text(
+    raw_input: Sequence[str],
+    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
+    pytorch_vocab: Vocab,
+) -> Sequence[int]: ...
+
+
+def encode_text(
+    raw_input: Union[str, Sequence[str]],
+    tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
+    pytorch_vocab: Vocab,
+) -> Sequence[int]:
+    """
+    TODO: Restructure this and related logic to not operate on union types.
+    """
+    input_tokenized = tokenizer(raw_input)  # type: ignore
+    input_as_ids = pytorch_vocab(input_tokenized)
+    return input_as_ids
+
+
 def get_tokenizer_encode_func(
     tokenizer: TokenizerProtocolRaw | TokenizerProtocolPreSplit,
     pytorch_vocab: Vocab,
 ) -> EncodeFuncProtocol:
-    @overload
-    def _encode_func(raw_input: str) -> Sequence[int]: ...
-
-    @overload
-    def _encode_func(raw_input: Sequence[str]) -> Sequence[int]: ...
-
-    def _encode_func(raw_input):
-        input_tokenized = tokenizer(raw_input)
-        input_as_ids = pytorch_vocab(input_tokenized)
-        return input_as_ids
-
-    return _encode_func
+    return partial(encode_text, tokenizer=tokenizer, pytorch_vocab=pytorch_vocab)
 
 
 def get_tokenized_vocab_iterator(
@@ -649,14 +733,20 @@ def yield_tokens_from_source(
     elif data_source_path.is_dir():
         iterator = tqdm(Path(data_source).iterdir(), desc="Vocabulary Setup")
         for file in iterator:
+            preserve_full = split_on is None
             yield from yield_tokens_from_file(
-                file_path=str(file), split_on=split_on, gathered_stats=gathered_stats
+                file_path=str(file),
+                split_on=split_on,
+                gathered_stats=gathered_stats,
+                preserve_full_content=preserve_full,
             )
 
     elif data_source_path.is_file():
         assert data_source_path.suffix == ".csv"
         yield from yield_tokens_from_csv(
-            file_path=data_source, split_on=split_on, gathered_stats=gathered_stats
+            file_path=data_source,
+            split_on=split_on,
+            gathered_stats=gathered_stats,
         )
 
     return gathered_stats
@@ -670,13 +760,14 @@ def yield_tokens_from_deeplake_dataset(
 ) -> Generator[Sequence[str], None, None]:
     deeplake_ds = load_deeplake_dataset(data_source=str(data_source))
     deeplake_iter = get_deeplake_input_source_iterable(
-        deeplake_dataset=deeplake_ds, inner_key=inner_key
+        deeplake_dataset=deeplake_ds,
+        inner_key=inner_key,
     )
 
     split_func = get_sequence_split_function(split_on=split_on)
 
     for sample in deeplake_iter:
-        cur_sequence = sample.text()
+        cur_sequence = str(sample)
         cur_line = split_func(cur_sequence)
 
         cur_length = len(cur_line)
@@ -692,9 +783,20 @@ def yield_tokens_from_file(
     file_path: str,
     split_on: Optional[str],
     gathered_stats: GatheredSequenceStats,
+    preserve_full_content: bool = False,
 ) -> Generator[Sequence[str], None, None]:
     gathered_stats.total_files += 1
 
+    if preserve_full_content:
+        with open(file_path, "r") as f:
+            content = f.read()
+            gathered_stats.total_count += 1
+            cur_length = len(content)
+            if cur_length > gathered_stats.max_length:
+                gathered_stats.max_length = cur_length
+            yield [content]
+        return
+
     split_func = get_sequence_split_function(split_on=split_on)
 
     with open(file_path, "r") as f:
