@@ -1,7 +1,7 @@
 from dataclasses import dataclass, field
 from typing import Literal, Sequence, Union
 
-al_cat_loss_names = Literal["CrossEntropyLoss"]
+al_cat_loss_names = Literal["CrossEntropyLoss", "BCEWithLogitsLoss"]
 al_con_loss_names = Literal[
     "MSELoss",
     "L1Loss",
@@ -27,6 +27,13 @@ class TabularOutputTypeConfig:
     :param cat_label_smoothing:
         Label smoothing to apply to categorical targets.
 
+    :param cat_loss_name:
+        Loss function to use for categorical targets. If using `BCEWithLogitsLoss`,
+        the targets should all be binary.
+
+    :param con_loss_name:
+        Loss function to use for continuous targets.
+
     :param uncertainty_weighted_mt_loss:
         Whether to use uncertainty weighted loss for multitask / multilabel learning.
     """
@@ -37,4 +44,4 @@ class TabularOutputTypeConfig:
     cat_label_smoothing: float = 0.0
     cat_loss_name: al_cat_loss_names = "CrossEntropyLoss"
     con_loss_name: al_con_loss_names = "MSELoss"
-    uncertainty_weighted_mt_loss: bool = True
+    uncertainty_weighted_mt_loss: bool = False
