@@ -1,13 +1,20 @@
 from dataclasses import dataclass
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Sequence
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    List,
+    Optional,
+    Sequence,
+)
 
-import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 import torch
 from ignite.engine import Engine
-from lifelines import KaplanMeierFitter
 from sklearn.preprocessing import LabelEncoder, StandardScaler
 
 from eir.data_load.label_setup import al_label_transformers_object
@@ -23,10 +30,13 @@ from eir.setup.output_setup import (
 from eir.setup.schemas import GlobalConfig
 from eir.target_setup.target_label_setup import MissingTargetsInfo
 from eir.train_utils import metrics, utils
-from eir.train_utils.evaluation_handlers.train_handlers_array_output import (
+from eir.train_utils.evaluation_modules.evaluation_output_survival import (
+    save_survival_evaluation_results_wrapper,
+)
+from eir.train_utils.evaluation_modules.train_handlers_array_output import (
     array_out_single_sample_evaluation_wrapper,
 )
-from eir.train_utils.evaluation_handlers.train_handlers_sequence_output import (
+from eir.train_utils.evaluation_modules.train_handlers_sequence_output import (
     sequence_out_single_sample_evaluation_wrapper,
 )
 from eir.train_utils.latent_analysis import (
@@ -128,11 +138,18 @@ def validation_handler(engine: Engine, handler_config: "HandlerConfig") -> None:
             val_ids=evaluation_results.ids_per_output,
             iteration=iteration,
             experiment=handler_config.experiment,
+            evaluation_metrics=evaluation_results.metrics_with_averages,
         )
 
     exp.model.train()
 
 
+def expand_binary_logits(binary_logits: torch.Tensor) -> torch.Tensor:
+    neg_logits = -binary_logits / 2
+    pos_logits = binary_logits / 2
+    return torch.cat([neg_logits, pos_logits], dim=-1)
+
+
 def run_all_eval_hook_analysis(
     hook_outputs: dict[str, LatentHookOutput],
     run_folder: Path,
@@ -282,7 +299,8 @@ def run_split_evaluation(
         output_objects=output_objects
     )
     split_output_generator = get_split_output_generator(
-        output_generator=output_generator, output_objects=output_objects
+        output_generator=output_generator,
+        output_objects=output_objects,
     )
 
     for split_outputs in split_output_generator:
@@ -299,13 +317,24 @@ def run_split_evaluation(
         full_gathered_ids_total += split_outputs.ids
 
         if with_labels and len(split_output_objects.compute) > 0:
+
+            filtered = metrics.filter_missing_outputs_and_labels(
+                batch_ids=split_outputs.ids,
+                model_outputs=output_compute,
+                target_labels=target_compute,
+                missing_ids_info=missing_ids_per_output,
+                with_labels=with_labels,
+            )
+
             cur_compute_metrics = compute_eval_metrics_wrapper(
-                val_outputs=output_compute,
-                val_target_labels=target_compute,
-                outputs=split_output_objects.compute,
+                val_outputs=filtered.model_outputs,
+                val_target_labels=filtered.target_labels,
+                output_objects=split_output_objects.compute,
                 experiment_metrics=experiment_metrics,
                 loss_function=loss_function,
                 device=device,
+                val_outputs_for_loss=output_compute,
+                val_targets_for_loss=target_compute,
             )
             batch_computed_metrics.append(cur_compute_metrics)
             batch_sizes.append(len(split_outputs.ids))
@@ -324,7 +353,7 @@ def run_split_evaluation(
         list_of_target_batch_dicts=full_gathered_output_batches_total
     )
 
-    filtered_values = metrics.filter_missing_outputs_and_labels(
+    filtered = metrics.filter_missing_outputs_and_labels(
         batch_ids=full_gathered_ids_total,
         model_outputs=full_gathered_output_batches_stacked,
         target_labels=full_gathered_label_batches_stacked,
@@ -335,27 +364,52 @@ def run_split_evaluation(
     full_gathered_metrics = {}
     if with_labels and len(split_output_objects.gather) > 0:
         full_gathered_metrics = compute_eval_metrics_wrapper(
-            val_outputs=filtered_values.model_outputs,
-            val_target_labels=filtered_values.target_labels,
-            outputs=split_output_objects.gather,
+            val_outputs=filtered.model_outputs,
+            val_target_labels=filtered.target_labels,
+            output_objects=split_output_objects.gather,
             experiment_metrics=experiment_metrics,
             loss_function=loss_function,
             device=device,
+            val_outputs_for_loss=full_gathered_output_batches_stacked,
+            val_targets_for_loss=full_gathered_label_batches_stacked,
         )
 
     all_metrics = {**computed_metrics_averaged, **full_gathered_metrics}
 
+    ids_per_output = _build_ids_per_output(filtered_values=filtered)
+
     evaluation_results = EvaluationResults(
         metrics_with_averages=all_metrics,
-        gathered_outputs=filtered_values.model_outputs,
-        gathered_labels=filtered_values.target_labels,
-        ids_per_output=filtered_values.ids,
+        gathered_outputs=filtered.model_outputs,
+        gathered_labels=filtered.target_labels,
+        ids_per_output=ids_per_output,
         all_ids=full_gathered_ids_total,
     )
 
     return evaluation_results
 
 
+def _build_ids_per_output(
+    filtered_values: metrics.FilteredOutputsAndLabels,
+) -> dict[str, dict[str, list[str]]]:
+    fv = filtered_values
+    ids_per_output = fv.ids
+    if ids_per_output is None:
+        assert fv.common_ids is not None
+        ids_per_output = {}
+        for output_name, inner_dict in fv.model_outputs.items():
+            if output_name not in ids_per_output:
+                ids_per_output[output_name] = {}
+
+            for inner_target_name in inner_dict.keys():
+                ids_per_output[output_name][inner_target_name] = fv.common_ids
+    else:
+        assert fv.common_ids is None
+        return ids_per_output
+
+    return ids_per_output
+
+
 def average_nested_dict_values(data: list[dict[str, Any]], weights: list[int]) -> dict:
     df = pd.json_normalize(data=data)
 
@@ -428,27 +482,40 @@ def get_split_outputs_map(output_objects: al_output_objects_as_dict) -> dict[str
 def compute_eval_metrics_wrapper(
     val_outputs: dict[str, dict[str, torch.Tensor]],
     val_target_labels: "al_training_labels_target",
-    outputs: al_output_objects_as_dict,
+    output_objects: al_output_objects_as_dict,
     experiment_metrics: metrics.al_metric_record_dict,
     loss_function: Callable,
     device: str,
+    val_outputs_for_loss: Optional[dict[str, dict[str, torch.Tensor]]] = None,
+    val_targets_for_loss: Optional["al_training_labels_target"] = None,
 ) -> metrics.al_step_metric_dict:
+    """
+    We have these specific optional values `val_outputs_for_loss` and
+    `val_targets_for_loss` as the loss computations assume we have equally
+    sized batches and they handle the NaNs themselves.
+
+    Later if we add NaN handling directly to metric calculations, we
+    can simplify this again to only accept `val_outputs` and `val_target_labels`.
+    """
     val_target_labels = model_training_utils.parse_tabular_target_labels(
-        output_objects=outputs, device=device, labels=val_target_labels
+        output_objects=output_objects, device=device, labels=val_target_labels
     )
 
     eval_metrics_dict = metrics.calculate_batch_metrics(
-        outputs_as_dict=outputs,
+        outputs_as_dict=output_objects,
         outputs=val_outputs,
         labels=val_target_labels,
         mode="val",
         metric_record_dict=experiment_metrics,
     )
 
-    val_losses = loss_function(inputs=val_outputs, targets=val_target_labels)
+    outputs_for_loss = val_outputs_for_loss or val_outputs
+    targets_for_loss = val_targets_for_loss or val_target_labels
+
+    val_losses = loss_function(inputs=outputs_for_loss, targets=targets_for_loss)
     val_loss_avg = metrics.aggregate_losses(losses_dict=val_losses)
     eval_metrics_dict_w_loss = metrics.add_loss_to_metrics(
-        outputs_as_dict=outputs,
+        outputs_as_dict=output_objects,
         losses=val_losses,
         metric_dict=eval_metrics_dict,
     )
@@ -457,7 +524,7 @@ def compute_eval_metrics_wrapper(
     assert isinstance(averaging_functions, dict)
     eval_metrics_dict_w_averages = metrics.add_multi_task_average_metrics(
         batch_metrics_dict=eval_metrics_dict_w_loss,
-        outputs_as_dict=outputs,
+        outputs_as_dict=output_objects,
         loss=val_loss_avg.item(),
         performance_average_functions=averaging_functions,
     )
@@ -465,133 +532,6 @@ def compute_eval_metrics_wrapper(
     return eval_metrics_dict_w_averages
 
 
-def save_survival_evaluation_results_wrapper(
-    val_outputs: dict[str, dict[str, torch.Tensor]],
-    val_labels: dict[str, dict[str, torch.Tensor]],
-    val_ids: dict[str, dict[str, list[str]]],
-    iteration: int,
-    experiment: "Experiment",
-) -> None:
-    for output_name, output_object in experiment.outputs.items():
-        output_type = output_object.output_config.output_info.output_type
-        if output_type != "survival":
-            continue
-
-        assert isinstance(output_object, ComputedSurvivalOutputInfo)
-
-        output_type_info = output_object.output_config.output_type_info
-        time_name = output_type_info.time_column
-        event_name = output_type_info.event_column
-
-        transformers = output_object.target_transformers
-        time_kbins_transformer = transformers[time_name]
-        time_bins = time_kbins_transformer.bin_edges_[0]
-
-        cur_sample_output_folder = utils.prepare_sample_output_folder(
-            output_folder=experiment.configs.gc.be.output_folder,
-            column_name=event_name,
-            output_name=output_name,
-            iteration=iteration,
-        )
-
-        ids = val_ids[output_name][event_name]
-        hazards_logits = val_outputs[output_name][event_name].cpu()
-        times_binned = val_labels[output_name][time_name].cpu().numpy()
-        events = val_labels[output_name][event_name].cpu().numpy().astype(int)
-
-        times = time_kbins_transformer.inverse_transform(times_binned.reshape(-1, 1))
-        times = times.flatten()
-
-        hazards = torch.sigmoid(hazards_logits).numpy()
-        survival_probs = np.cumprod(1 - hazards, 1)
-
-        time_bins_except_last = time_bins[:-1]
-        plot_survival_curves(
-            times=times,
-            events=events,
-            predicted_probs=survival_probs,
-            time_bins=time_bins_except_last,
-            output_folder=cur_sample_output_folder,
-        )
-
-        event_transformer = transformers[event_name]
-        events_untransformed = event_transformer.inverse_transform(events)
-
-        df = pd.DataFrame(
-            {
-                "ID": ids,
-                time_name: times,
-                event_name: events,
-                f"{event_name} Untransformed": events_untransformed,
-                "Predicted_Risk": hazards[:, -1],
-            }
-        )
-
-        for i, t in enumerate(time_bins_except_last):
-            df[f"Surv_Prob_t{i}"] = survival_probs[:, i]
-
-        csv_path = f"{cur_sample_output_folder}/survival_predictions.csv"
-        df.to_csv(csv_path, index=False)
-
-        plot_individual_survival_curves(
-            df=df,
-            time_bins=time_bins_except_last,
-            output_folder=str(cur_sample_output_folder),
-            n_samples=5,
-        )
-
-
-def plot_survival_curves(
-    times,
-    events,
-    predicted_probs,
-    time_bins,
-    output_folder,
-):
-    kmf = KaplanMeierFitter()
-    kmf.fit(times, events)
-
-    plt.figure(figsize=(10, 6))
-    kmf.plot_survival_function(label="Kaplan-Meier Estimate")
-
-    mean_predicted = predicted_probs.mean(axis=0)
-    plt.step(time_bins, mean_predicted, where="post", label="Mean Predicted")
-
-    plt.xlabel("Time")
-    plt.ylabel("Survival Probability")
-    plt.legend()
-    plt.title("Kaplan-Meier vs Mean Predicted Survival")
-    plt.savefig(f"{output_folder}/survival_curves.png")
-    plt.close()
-
-
-def plot_individual_survival_curves(
-    df: pd.DataFrame,
-    time_bins: np.ndarray,
-    output_folder: str,
-    n_samples: int = 5,
-) -> None:
-    random_samples = df.sample(n=n_samples)
-
-    plt.figure(figsize=(12, 8))
-
-    for _, row in random_samples.iterrows():
-        surv_probs = row[
-            [col for col in df.columns if col.startswith("Surv_Prob_t")]
-        ].to_numpy()
-
-        plt.step(x=time_bins, y=surv_probs, where="post", label=f"ID: {row['ID']}")
-
-    plt.xlabel("Time")
-    plt.ylabel("Survival Probability")
-    plt.title(f"Survival Curves for {n_samples} Random Samples")
-    plt.legend()
-    plt.grid(True, alpha=0.3)
-
-    plt.savefig(f"{output_folder}/individual_survival_curves.pdf")
-    plt.close()
-
-
 def save_tabular_evaluation_results_wrapper(
     val_outputs: Dict[str, Dict[str, torch.Tensor]],
     val_labels: Dict[str, Dict[str, torch.Tensor]],
@@ -614,15 +554,35 @@ def save_tabular_evaluation_results_wrapper(
                     output_name=output_name,
                     iteration=iteration,
                 )
-                cur_val_outputs = val_outputs[output_name][column_name]
-                cur_val_outputs_np = cur_val_outputs.cpu().numpy()
 
-                cur_val_labels = val_labels[output_name][column_name]
-                cur_val_labels_np = cur_val_labels.cpu().numpy()
-                target_transformers = output_object.target_transformers
+                cur_val_outputs = val_outputs[output_name][column_name]
+
+                if column_type == "cat" and cur_val_outputs.shape[1] == 1:
+                    cur_val_outputs = expand_binary_logits(
+                        binary_logits=cur_val_outputs
+                    )
 
                 cur_val_ids = val_ids[output_name][column_name]
 
+                cur_val_labels = val_labels[output_name][column_name]
+                filtered = metrics.filter_tabular_missing_targets(
+                    outputs=cur_val_outputs,
+                    target_labels=cur_val_labels,
+                    ids=cur_val_ids,
+                    target_type=column_type,
+                )
+
+                to_npy = metrics.general_torch_to_numpy
+                cur_val_outputs_np = to_npy(tensor=filtered.model_outputs)
+
+                cur_val_labels_np = to_npy(tensor=filtered.target_labels)
+                if column_type == "cat":
+                    cur_val_labels_np = cur_val_labels_np.astype(int)
+
+                cur_val_ids = filtered.ids
+
+                target_transformers = output_object.target_transformers
+
                 plot_config = PerformancePlotConfig(
                     val_outputs=cur_val_outputs_np,
                     val_labels=cur_val_labels_np,
@@ -654,20 +614,24 @@ def save_evaluation_results(
 ) -> None:
     pc = plot_config
 
-    common_args = {
-        "val_outputs": pc.val_outputs,
-        "val_labels": pc.val_labels,
-        "val_ids": pc.val_ids,
-        "output_folder": pc.output_folder,
-        "transformer": pc.target_transformer,
-    }
-
     vf.gen_eval_graphs(plot_config=pc)
 
     if pc.column_type == "cat":
-        save_classification_predictions(**common_args)
+        save_classification_predictions(
+            val_labels=pc.val_labels,
+            val_outputs=pc.val_outputs,
+            val_ids=pc.val_ids,
+            transformer=pc.target_transformer,
+            output_folder=pc.output_folder,
+        )
     elif pc.column_type == "con":
-        scale_and_save_regression_predictions(**common_args)
+        scale_and_save_regression_predictions(
+            val_labels=pc.val_labels,
+            val_outputs=pc.val_outputs,
+            val_ids=pc.val_ids,
+            transformer=pc.target_transformer,
+            output_folder=pc.output_folder,
+        )
 
 
 def save_classification_predictions(
@@ -677,6 +641,7 @@ def save_classification_predictions(
     transformer: LabelEncoder,
     output_folder: Path,
 ) -> None:
+
     validation_predictions_total = val_outputs.argmax(axis=1)
 
     df_predictions = _parse_valid_classification_predictions(
