@@ -468,7 +468,8 @@ class MyRunningAverage(RunningAverage):
 
 
 def _attach_running_average_metrics(
-    engine: Engine, monitoring_metrics: List[Tuple[str, str, str]]
+    engine: Engine,
+    monitoring_metrics: List[Tuple[str, str, str]],
 ) -> None:
     """
     For each metric, we create an output_transform function that grabs the
@@ -488,10 +489,10 @@ def _attach_running_average_metrics(
             column_name_key: str,
             metric_name_key: str,
         ) -> float:
-            value = metric_dict_from_step[output_name_key][column_name_key][
-                metric_name_key
-            ]
-            return value
+            output_dict = metric_dict_from_step[output_name_key]
+            inner_dict = output_dict[column_name_key]
+            metric_value = inner_dict[metric_name_key]
+            return metric_value
 
         partial_func = partial(
             output_transform,
@@ -624,14 +625,6 @@ def _attach_run_event_handlers(trainer: Engine, handler_config: HandlerConfig):
             handler_config=handler_config,
         )
 
-    if exp.hooks.custom_handler_attachers is not None:
-        custom_handlers = _get_custom_handlers(handler_config=handler_config)
-        trainer = _attach_custom_handlers(
-            trainer=trainer,
-            handler_config=handler_config,
-            custom_handlers=custom_handlers,
-        )
-
     return trainer
 
 
@@ -901,21 +894,3 @@ def _plot_progress_handler(engine: Engine, handler_config: HandlerConfig) -> Non
         title_extra="Multi Task Average",
         plot_skip_steps=gc.vl.plot_skip_steps,
     )
-
-
-def _get_custom_handlers(handler_config: "HandlerConfig"):
-    custom_handlers = handler_config.experiment.hooks.custom_handler_attachers
-
-    return custom_handlers
-
-
-def _attach_custom_handlers(
-    trainer: Engine, handler_config: "HandlerConfig", custom_handlers
-):
-    if not custom_handlers:
-        return trainer
-
-    for custom_handler_attacher in custom_handlers:
-        trainer = custom_handler_attacher(trainer, handler_config)
-
-    return trainer
